//	Teeminator 1.0
//
//	Copyright (C) 2010 Paul Wilhelm <paul@mosfetkiller.de>
//	http://mosfetkiller.de/?s=teeminator

#include "main.h"

//	Prototypen
void init();
void update_LEDs();
void power_off();
void servo_on();
void servo_off();

//	Variablen

//	Zeit
volatile unsigned int my_timer_0 = 0;

//	Servopositionen: 0 < pos < 200
volatile unsigned char SERVO_MAIN_pos = 140;		//	140 = Grundstellung (waagerecht)
volatile char SERVO_MAIN_offset = 0;

volatile unsigned char SERVO_SQUEEZE_pos = 25;	//	25 = Grundstellung, 40 = Angriffsstellung, 78 = Quetschen

//	LEDs
char LED_STATUS_state = FALSE;
unsigned char DIE_count;

//	Main
int main(void)
{
	//	Initialisieren
	init();

	//	Würfelwert aus EEPROM laden
	DIE_count = eeprom_read_byte(0);

	update_LEDs();

	//	Servos positionieren lassen
	my_timer_0 = 0; while (my_timer_0 < 40);
	servo_off();

	//	Variablen für Encoderauswertung
	unsigned char old_state_a, new_state_a, old_state_b, new_state_b,	encdec_first = TRUE;
	unsigned char rising_edge_a = FALSE, falling_edge_a = FALSE, rising_edge_b = FALSE, falling_edge_b = FALSE;

	//	Hauptschleife
	while(1)
	{
		//	Servooffset kann per Poti eingestellt werden
		char SERVO_MAIN_offset_old;
		set_bit(ADCSR, ADSC);
		loop_until_bit_is_clear(ADCSR, ADSC);
		SERVO_MAIN_offset = (char)(ADCH / 6);

		//	Servo bei Offsetänderung neu positionieren
		//	Servos für kurze Zeit eingeschaltet lassen, damit es nicht so ruckelt
		if (SERVO_MAIN_offset != SERVO_MAIN_offset_old)
		{
			SERVO_MAIN_offset_old = SERVO_MAIN_offset;
			servo_on();
			my_timer_0 = 0;
		}

		if (bit_is_set(SERVO_MOSFET_port, SERVO_MOSFET_bit) && my_timer_0 >= 75) servo_off();


		//	Encoderauswertung
		new_state_a = !bit_is_set(ENC_A_pin, ENC_A_bit);
		new_state_b = !bit_is_set(ENC_B_pin, ENC_B_bit);
		if (encdec_first) { old_state_a = new_state_a; old_state_b = new_state_b; encdec_first = FALSE; }

		if (new_state_a != old_state_a)
		{
			old_state_a = new_state_a;
			if (new_state_a) rising_edge_a = TRUE; else falling_edge_a = TRUE;
		} else { rising_edge_a = FALSE; falling_edge_a = FALSE; }

		if (new_state_b != old_state_b)
		{
			old_state_b = new_state_b;
			if (new_state_b) rising_edge_b = TRUE; else falling_edge_b = TRUE;
		} else { rising_edge_b = FALSE; falling_edge_b = FALSE; }


		//	Der Würfelwert kann mit dem Encoder verändert werden
		if ((falling_edge_a && new_state_b) || (rising_edge_a && !new_state_b))
		{
			if (DIE_count < 7) DIE_count++; else DIE_count = 1;
			update_LEDs();
		}

		if ((rising_edge_a && new_state_b) || (falling_edge_a && !new_state_b))
		{
			if (DIE_count > 1) DIE_count--; else DIE_count = 7;
			update_LEDs();
		}

		//	Sequenz bei Tasterdruck beginnen
		if (!bit_is_set(BUTTON_pin, BUTTON_bit))
		{
			//	Würfelwert in EEPROM speichern
			eeprom_write_byte(0, DIE_count);

			//	Status-LED einschalten
			LED_STATUS_state = TRUE;
			update_LEDs();

			//	Servos aktivieren
			servo_on();

			//	Beutel in Tasse absenken
			for (SERVO_MAIN_pos = 140; SERVO_MAIN_pos < (200 - SERVO_MAIN_offset); SERVO_MAIN_pos++) _delay_us(250);

			//	Servos deaktivieren
			servo_off();

			//	Die eingestellte Zeit (in Minuten) abwarten,
			//	alle 30 s den Beutel kurz herausziehen und wieder eintauchen
			//	Vorgang kann mit Tasterdruck abgebrochen werden
			//	Verbleibende Zeit auf Würfel anzeigen
			char canceled = FALSE;
			unsigned char DIE_count_copy = DIE_count;
			my_timer_0 = 0;
			while (my_timer_0 < DIE_count_copy * 60 * 50 && !canceled)
			{
				if (my_timer_0 % (30 * 50) == 0 && my_timer_0 > 1)
				{
					//	Servos aktivieren
					servo_on();

					//	Beutel herausziehen und wieder eintauchen
					for (SERVO_MAIN_pos = 200 - SERVO_MAIN_offset; SERVO_MAIN_pos >= 165 - SERVO_MAIN_offset; SERVO_MAIN_pos--) _delay_us(250);
					_delay_ms(20);
					for (SERVO_MAIN_pos = 165 - SERVO_MAIN_offset; SERVO_MAIN_pos <= 200 - SERVO_MAIN_offset; SERVO_MAIN_pos++) _delay_us(250);

					//	Servps deaktivieren
					servo_off();
				}

				// Verbleibende Zeit anzeigen
				DIE_count = DIE_count_copy - my_timer_0 / (60*50);
				update_LEDs();

				if (!bit_is_set(BUTTON_pin, BUTTON_bit)) canceled = TRUE;
			}

			//	Servos aktivieren
			servo_on();

			if (!canceled)
			{
				DIE_count = 0;
				update_LEDs();

				//	Bautel aus dem Wasser ziehen
				SERVO_SQUEEZE_pos = 55;
				for (SERVO_MAIN_pos = 200 - SERVO_MAIN_offset; SERVO_MAIN_pos >= 160 - SERVO_MAIN_offset; SERVO_MAIN_pos--) _delay_us(250);

				//	Beutel mehrmals ausquetschen und dabei immer etwas weiter hochziehen
				for (unsigned char i = 0; i < 5; i++)
				{
					if (i > 0)
					{
						my_timer_0 = 0; while (my_timer_0 < 25);
					}
					SERVO_SQUEEZE_pos = 78;

					SERVO_MAIN_pos -= 5;

					my_timer_0 = 0; while (my_timer_0 < 50);
					if (i < 4)
					{
						SERVO_SQUEEZE_pos = 40;
					} else SERVO_SQUEEZE_pos = 25;
				}

				char SERVO_MAIN_offset_copy = SERVO_MAIN_offset; SERVO_MAIN_offset = 0;

				//	Arm zum Mülleimer schwenken
				for (SERVO_MAIN_pos; SERVO_MAIN_pos > 0; SERVO_MAIN_pos--) _delay_us(250);

				//	Beuteletikett mehrmals versuchen abzuschütteln

				for (unsigned char i = 0; i < 5; i++)
				{
					SERVO_MAIN_pos = 30; my_timer_0 = 0; while (my_timer_0 < 10);
					SERVO_MAIN_pos = 0; my_timer_0 = 0; while (my_timer_0 < 30);
				}

				SERVO_MAIN_offset = SERVO_MAIN_offset_copy;

				//	Kurz warten; letzte Chance für den Benutzer, den Beutel manuell zu entfernen
				my_timer_0 = 0; while (my_timer_0 < 175);

				//	Arm wieder in Grundstellung fahren
				SERVO_MAIN_pos = 140;
			} else {
				//	Vorgang wurde abgebrochen; Arm langsam in Grundstellung fahren
				for (SERVO_MAIN_pos; SERVO_MAIN_pos >= 140; SERVO_MAIN_pos--) _delay_us(250);
			}

			//	Kurz warten, bis sich der Hauptservo positioniert hat
			my_timer_0 = 0; while (my_timer_0 < 40);

			//	Würfelwert aus EEPROM laden
			DIE_count = eeprom_read_byte(0);

			//	Status-LED ausschalten
			LED_STATUS_state = FALSE;

			//	Servos deaktivieren
			servo_off();

			//	AVR schlafen legen
			power_off();

			update_LEDs();

			//	Servos positionieren lassen
			servo_on();
			my_timer_0 = 0; while (my_timer_0 < 40);
			servo_off();
		}
	}
}


//	Initialisieren
void init()
{
	//	Ports vorbereiten

	//	Pullups an Encoderausgängen
	set_bit(ENC_A_port, ENC_A_bit);
	set_bit(ENC_B_port, ENC_B_bit);

	//	Servos
	set_bit(SERVO_MAIN_ddr, SERVO_MAIN_bit);
	set_bit(SERVO_SQUEEZE_ddr, SERVO_SQUEEZE_bit);

	//	Würfel
	set_bit(DIE_1_ddr, DIE_1_bit);
	set_bit(DIE_2_ddr, DIE_2_bit);
	set_bit(DIE_3_ddr, DIE_3_bit);
	set_bit(DIE_4_ddr, DIE_4_bit);

	//	Pullup an Taster (Encoder)
	set_bit(BUTTON_port, BUTTON_bit);

	//	Status-LED
	set_bit(LED_STATUS_ddr, LED_STATUS_bit);

	//	Poti-Stromversorgung einschalten
	set_bit(POTPWR_ddr, POTPWR_bit);
	set_bit(POTPWR_port, POTPWR_bit);

	set_bit(SERVO_MOSFET_ddr, SERVO_MOSFET_bit);

	//	PB2 = Output; ansonsten würde dieser Pin unkontrolliert einen PCINT0 auslösen
	set_bit(DDRB, PB2);

	cli();

	//	ADC
	ADCSR = (1 << ADEN);				//	ADC Enable
	ADMUX = POTI_adc | (1 << ADLAR);	//	ADLAR

	//	TIMER1 (Servo, Zeit)
	TCCR1B = (1 << CTC1) | (1 << CS11) | (1 << CS10);	//	CTC-Mode, Prescaler 8
	OCR1C = 24;											//	Periodendauer 10 µS
	TIMSK = (1 << OCIE1A);

	sei();
}


void update_LEDs()
{
	//	LEDs steuern

	//	Status-LED
	if (LED_STATUS_state) set_bit(LED_STATUS_port, LED_STATUS_bit); else clear_bit(LED_STATUS_port, LED_STATUS_bit);

	//	Würfel
	if (DIE_count % 2 == 1) set_bit(DIE_1_port, DIE_1_bit); else clear_bit(DIE_1_port, DIE_1_bit);
	if (DIE_count > 2) set_bit(DIE_2_port, DIE_2_bit); else clear_bit(DIE_2_port, DIE_2_bit);
	if (DIE_count > 5) set_bit(DIE_3_port, DIE_3_bit); else clear_bit(DIE_3_port, DIE_3_bit);
	if (DIE_count == 2 || DIE_count > 3) set_bit(DIE_4_port, DIE_4_bit); else clear_bit(DIE_4_port, DIE_4_bit);
}



//	Abschalten (Schlafmodus)
void power_off()
{
	//	Poti-Stromversorgung ausschalten
	clear_bit(POTPWR_port, POTPWR_bit);

	//	LEDs aus

	//	Status-LED
	clear_bit(LED_STATUS_port, LED_STATUS_bit);

	//	Würfel
	clear_bit(DIE_1_port, DIE_1_bit);
	clear_bit(DIE_2_port, DIE_2_bit);
	clear_bit(DIE_3_port, DIE_3_bit);
	clear_bit(DIE_4_port, DIE_4_bit);

	//	Pullups an Encoderausgängen ausschalten
	clear_bit(ENC_A_port, ENC_A_bit);
	clear_bit(ENC_B_port, ENC_B_bit);

	//	TIMER1 abschalten
	cli();
	TIMSK = 0;
	ADCSR = 0;				//	ADC Disable

	//	PCINT0 aktivieren
	set_bit(GIMSK, PCIE0);

	sei();

	//	PCIF löschen
	set_bit(GIFR, PCIF);

	//	Schlafen, bis der Taster gedrückt wird
	set_bit(MCUCR, SM1);	//	Power-down Mode
	clear_bit(MCUCR, SM0);
	set_bit(MCUCR, SE);
	asm volatile("sleep"::);

	//	Hier geht es nach dem Aufwachen dann wieder weiter

	//	TIMER1 aktivieren
	cli();
	TIMSK = (1 << OCIE1A);
	ADCSR = (1 << ADEN);				//	ADC Enable
	//	PCINT0 deaktivieren
	clear_bit(GIMSK, PCIE0);
	sei();

	//	Poti-Stromversorgung einschalten
	set_bit(POTPWR_port, POTPWR_bit);

	//	Pullups an Encoderausgängen einschalten
	set_bit(ENC_A_port, ENC_A_bit);
	set_bit(ENC_B_port, ENC_B_bit);

	//	PCIF löschen
	set_bit(GIFR, PCIF);
}


//	Dient nur zum Aufwachen
ISR(IO_PINS_vect)
{
}


//	Interruptvektor von TIMER1 (Compare A) (Servo, Zeit)
//	*** Diese ISR ist bereits sehr vollgepackt, man kann im Grunde keine Funktionen mehr hinzufügen ***
ISR(TIMER1_CMPA_vect)
{
	static int pwm_counter;

	//	Die komischen Servos brauchen einen Offset von 45 (keine Ahnung warum...)


	//	Hauptservo
	if (pwm_counter < (100 + (SERVO_MAIN_pos + SERVO_MAIN_offset) - 45))
		set_bit(SERVO_MAIN_port, SERVO_MAIN_bit);
	else
		clear_bit(SERVO_MAIN_port, SERVO_MAIN_bit);

	//	Quetschservo
	if (pwm_counter < 100 + SERVO_SQUEEZE_pos - 45)
		set_bit(SERVO_SQUEEZE_port, SERVO_SQUEEZE_bit);
	else
		clear_bit(SERVO_SQUEEZE_port, SERVO_SQUEEZE_bit);

	//	PWM-Counter
	if (pwm_counter < 2000)
	{
		pwm_counter++;
	} else {
		pwm_counter = 0;

		//	my_timer_0 hat beim Starten eine max. Abweichung von 20 ms
		my_timer_0++;
	}
}


void servo_on()
{
	set_bit(SERVO_MOSFET_port, SERVO_MOSFET_bit);
}

void servo_off()
{
	clear_bit(SERVO_MOSFET_port, SERVO_MOSFET_bit);

	//	Wenn diese Pins low sind, zieht der Servo immer noch Strom, auch wenn GND getrennt ist
	set_bit(SERVO_SQUEEZE_port, SERVO_SQUEEZE_bit);
	set_bit(SERVO_MAIN_port, SERVO_MAIN_bit);
}
